# -*- coding: utf-8 -*-
# ===============================================================
#  @author: huangchi@zhiyitech.cn
#  @date: 2018/8/29 上午8:57
#  @brief: 
# ===============================================================

import os

# 是否为测试/开发环境
TEST = int(os.getenv('TEST', 0))
# flag 标识
FLAG = os.getenv('FLAG')
# redis key 分隔符
KEY_DELIMITER = '___'
# 版本号的格式
DATE_FMT = '%Y/%m/%d'
VERSION_FMT = f'{DATE_FMT} %H:%M'
# 存放监控key的db
REDIS_DB = 14


class Sql:
    INSERT = 'insert into `{tb}` ({columns}) values ({fmt}) '
    INSERT_IGNORE = 'insert ignore into `{tb}` ({columns}) values ({fmt}) '


class TBName:
    TASK_MONITOR = 'task_monitor'


class RedisKeyPrefix:
    TASK_MONITOR = '(task_monitor)'


class RedisKey:
    # 不同任务配置信息的键
    TASK_CONFIG = f'{RedisKeyPrefix.TASK_MONITOR}hash_task_config_{{bis_name}}_{{task_name}}'


class EmailAddress:
    """邮箱地址"""
    ZEYU = 'zeyu@zhiyitech.cn'
    HUIHUI = 'huihui@zhiyitech.cn'
    XIANGYOU = 'liheyou@zhiyitech.cn'
    HUANGCHI = 'huangchi@zhiyitech.cn'



# -*- coding: utf-8 -*-
# ===============================================================
#  @author: huangchi@zhiyitech.cn
#  @date: 2018/9/17 上午10:43
#  @brief: 
# ===============================================================

# mysql连接池配置
sql_engine_config = {
    # 默认连接池数量
    'pool_size': 1,
    # 最多允许溢出的连接数
    'max_overflow': 4,
    # 连接池回收时间
    'pool_recycle': 5,
    # 与mysql交互的超时时间
    'pool_timeout': 15,
    # 在与mysql交互前先检验连接是否有效
    'pool_pre_ping': True
}

# 邮件服务器的配置
mail_config = {
    'host': 'smtp.exmail.qq.com',
    'port': 465,
    'user': 'alert@zhiyitech.cn',
    'password': 'BUpiEjgxJjNRKGAN'
}



# -*- coding: utf-8 -*-
# ===============================================================
#  @author: huangchi@zhiyitech.cn
#  @date: 2018/8/29 下午2:19
#  @brief: 监控
# ===============================================================

import time
import json

from ..common import constants, util
from ..core.mail import Mail

from ..core.task_dao import TaskDao

redis_key_prefixes = constants.RedisKeyPrefix
redis_keys = constants.RedisKey
key_delimiter = constants.KEY_DELIMITER


class DataResult:
    SUCCESS = 'success'
    FAILED = 'failed'
    RETRY = 'retry'


class Monitor:
    def __init__(self, bis_name, task_name, redis_config, mail_config):
        """
        初始化监控
        :param bis_name:        业务名
        :param task_name:       任务名
        :param redis_config:    redis的配置
        :param mail_config:     邮件的配置
        """
        if key_delimiter in bis_name or key_delimiter in task_name:
            raise Exception(f'业务名或任务名中不能包含{key_delimiter}')
        self.bis_name = bis_name
        self.task_name = task_name
        self.dao = TaskDao(redis_config, mysql_config=None)
        self.redis_client = self.dao.redis_client
        self.mail = Mail(mail_config, time_interval=5 * 60)
        self.task_config_key = self.cons_config_key()

    @staticmethod
    def cons_version():
        """构造当前时间的版本号"""
        return time.strftime(constants.VERSION_FMT, time.localtime())

    def cons_config_key(self):
        """构造配置key"""
        return redis_keys.TASK_CONFIG.format(bis_name=self.bis_name, task_name=self.task_name)

    def cons_monitor_key(self, version):
        """构造监控key"""
        # 检查版本号的格式是否正确
        time.strptime(version, constants.VERSION_FMT)
        # 构造监控key
        monitor_key = f'{redis_key_prefixes.TASK_MONITOR}hash{key_delimiter}{self.bis_name}{key_delimiter}' \
                      f'{self.task_name}{key_delimiter}{version}'
        return monitor_key

    def init_config(self, failure_rate_up: float=.0, avg_time_up: float=5.0, warn_addresses: list=None):
        """初始化任务的配置信息"""
        self.redis_client.hset(self.task_config_key, 'failure_rate_up', failure_rate_up)
        self.redis_client.hset(self.task_config_key, 'avg_time_up', avg_time_up)
        if warn_addresses:
            self.redis_client.hset(self.task_config_key, 'warn_addresses', json.dumps(warn_addresses))

    def create(self, version, queue_length: int):
        """初始化监控"""
        monitor_key = self.cons_monitor_key(version)
        if self.redis_client.hgetall(monitor_key):
            return
        # 初始化监控
        self.redis_client.hsetnx(monitor_key, 'bis_name', self.bis_name)
        self.redis_client.hsetnx(monitor_key, 'task_name', self.task_name)
        self.redis_client.hsetnx(monitor_key, 'version', version)
        self.redis_client.hsetnx(monitor_key, 'queue_length', queue_length)
        self.redis_client.hsetnx(monitor_key, 'left', queue_length)
        # 设置过期时间
        self.redis_client.expire(monitor_key, 2 * 24 * 3600)
        # 对上一版本的任务的报警
        self.warning(version)
        # 设置last_version
        self.redis_client.hset(self.task_config_key, 'last_version', version)

    def send(self, version, data: int=1, data_result=DataResult.SUCCESS, crawl_time=None):
        """
        发送监控信息
        :param version:
        :param data: 爬取的种子数，默认1
        :param data_result: 爬取结果，默认 success
        :param crawl_time: 爬取时间，默认None
        :return:
        """
        # 判断监控是否已初始化
        monitor_key = self.cons_monitor_key(version)
        if not self.redis_client.hgetall(monitor_key):
            raise Exception('Please init your monitor first!')
        # 发送监控信息
        self.redis_client.hincrby(monitor_key, data_result, data)
        if data_result == DataResult.SUCCESS or data_result == DataResult.FAILED:
            self.redis_client.hincrby(monitor_key, 'left', -data)
        if crawl_time:
            self.redis_client.hincrbyfloat(monitor_key, 'crawl_time', crawl_time)

    def incr_queue_length(self, version, queue_length: int):
        """更新队列长度"""
        monitor_key = self.cons_monitor_key(version)
        self.redis_client.hincrby(monitor_key, 'queue_length', queue_length)
        self.redis_client.hincrby(monitor_key, 'left', queue_length)

    def warning(self, version):
        """对上一版本的任务的报警"""
        # 获取任务的配置信息
        task_config = self.redis_client.hgetall(self.task_config_key)
        # 未配置任务的配置信息
        if not task_config:
            return
        # 未配置收件人
        if not task_config.get(b'warn_addresses'):
            return
        # 初次爬取
        if not task_config.get(b'last_version'):
            self.redis_client.hset(self.task_config_key, 'last_version', version)
            return
        # 获取上一个版本的任务信息
        last_version = task_config[b'last_version'].decode()
        monitor_key = self.cons_monitor_key(last_version)
        last_task = self.redis_client.hgetall(monitor_key)
        if not last_task:
            return
        content = util.task_detect(last_task, task_config)
        # 无报警信息
        if content == '':
            return
        # 报警
        subject = f'业务：{self.bis_name}, 任务：{self.task_name}, 版本：{last_version} 检测结果'
        warn_addresses = json.loads(task_config[b'warn_addresses'])
        self.mail.send(subject, content, addresses=warn_addresses)



# -*- coding: utf-8 -*-
# ===============================================================
#  @author: huangchi@zhiyitech.cn
#  @date: 2018/9/11 上午8:24
#  @brief: 
# ===============================================================

import json
import logging

from ..common import util, config

# 已初始化的redis连接池
redis_clients = dict()
# 已初始化的mysql连接池
mysql_clients = dict()


class RedisPosition:
    HEAD = 'head'
    TAIL = 'tail'


class Dao:
    """
    持久层基类
    包含 python - mysql，python - redis 的交互
    """
    def __init__(self, redis_config=None, mysql_config=None, redis_key=None):
        self.redis_key = redis_key
        self.redis_client = self._init_redis_client_(redis_config)
        self.mysql_client = self._init_mysql_client_(mysql_config)

    @staticmethod
    def _singleton_(_class, cls):
        """
        构造单例
        当需要将类型设置为单例模式时，重写当前类型的__new__方法
        并调用<code>Dao._singleton_(_class, cls)</code>, 获取返回值即可
        :param _class: 类的类型
        :param cls: 类的实例
        :return:
        """
        if not cls._instance:
            _class._instance = object.__new__(cls)
        return cls._instance

    @staticmethod
    def _init_redis_client_(redis_config):
        """获取redis连接池"""
        if not redis_config:
            return
        # 判断是否已初始化过该连接
        config_md5_code = util.get_dict_hash(redis_config)
        redis_client = redis_clients.get(config_md5_code)
        # 初始化redis连接
        if not redis_client:
            redis_client = util.get_redis_conn(redis_config)
            redis_clients.setdefault(config_md5_code, redis_client)
        return redis_client

    @staticmethod
    def _init_mysql_client_(mysql_config):
        """获取mysql连接池"""
        if not mysql_config:
            return
        # 判断是否已初始化过该连接
        config_md5_code = util.get_dict_hash(mysql_config)
        mysql_client = mysql_clients.get(config_md5_code)
        # 初始化mysql连接
        if not mysql_client:
            mysql_client = util.get_mysql_conn(mysql_config, config.sql_engine_config)
            mysql_clients.setdefault(config_md5_code, mysql_client)
        return mysql_client

    @util.sql_engine()
    def store(self, sql, tb_name, *data, session):
        """
        mysql数据存储
        :param sql: 预定义sql语句，一般为 lib.constants.Sql 中定义的sql语句
        :param tb_name: 数据表名
        :param data: 待存储的数据列表，元素类型为dict
        :param session: 数据库连接，由装饰器生成
        :return: 受影响的行数
        """
        logging.debug('store mysql start...')
        if len(data) == 0:
            return len(data)
        # 获取列
        columns = ','.join([f'`{key}`' for key in data[0].keys()])
        # 占位符
        fmt = ','.join([':' + key for key in data[0].keys()])
        # 格式化sql
        sql = sql.format(tb=tb_name, columns=columns, fmt=fmt)
        # 存储
        cursor = session.execute(sql, data)
        logging.debug('store mysql end...')
        return cursor.rowcount

    @util.sql_engine()
    def get(self, sql, session):
        """从mysql中获取数据，获取到的数据类型为dict"""
        logging.debug('query mysql start...')
        cursor = session.execute(sql)
        values = cursor.fetchall()
        values = [dict(value) for value in values]
        logging.debug(f'query mysql end, get {len(values)} records.')
        return values

    def get_seed(self, position=RedisPosition.HEAD):
        """
        从redis中获取一个种子
        若队列中无元素，则会无限期阻塞至拿到元素为止
        :param position:    拿取种子的位置，分两种
                            head（头）与 tail（尾）
                            'head' 移出并获取列表的第一个元素
                            'tail' 移除并获取列表最后一个元素
        :return: redis key中的一个元素，并进行 json.loads() 操作
        """
        seed = None
        if position == RedisPosition.HEAD:
            redis_key, seed = self.redis_client.blpop(self.redis_key)
        elif position == RedisPosition.TAIL:
            redis_key, seed = self.redis_client.brpop(self.redis_key)
        if seed:
            seed = json.loads(seed)
        return seed

    def push_seed(self, *seeds, position=RedisPosition.TAIL):
        """
        将种子放入队列中
        seeds 中的元素类型需为 str，若为其它类型（如dict）
        则在取出种子后 json.loads(seed) 时会出错
        :param seeds: 待存入的种子列表
        :param position: 插入位置，默认队尾（right push）
        :return: None
        """
        if len(seeds) == 0:
            return
        if position == RedisPosition.HEAD:
            self.redis_client.lpush(self.redis_key, *seeds)
        elif position == RedisPosition.TAIL:
            self.redis_client.rpush(self.redis_key, *seeds)

    def push_seed_from_mysql(self, sql):
        """从mysql中获取数据存入redis中"""
        data = self.get(sql)
        length = len(data)
        if length == 0:
            return
        data = [json.dumps(value) for value in data]
        self.push_seed(*data)
        return length
